class DS28C50Error(Exception):
    pass
    
class AuthenticationError(DS28C50Error):
    #0x00
    pass

class InternalError(DS28C50Error):
    #0x22
    pass
    
class InvalidSequenceError(DS28C50Error):
    #0x33
    pass

class PageProtectionError(DS28C50Error):
    #0x55
    pass
    
class InvalidParameterError(DS28C50Error):
    #0x77
    pass
    
class DeviceDisabledError(DS28C50Error):
    #0x88
    pass
    
class HealthTestError(DS28C50Error):
    #0xDD
    pass

class UnknownResultByteError(DS28C50Error):
    pass


def processResultByte(result_byte):
    if result_byte == 0x00:
        raise AuthenticationError("HMACs did not match. Please verify HMAC message is correct and the right secret is being used")
    elif result_byte == 0x22:
        raise InternalError("Internal device error")
    elif result_byte == 0x33:
        raise InvalidSequenceError("Make sure that proper steps where done. See users guide.")
    elif result_byte == 0x55:
        raise PageProtectionError("Page Protection prevents this command from running")
    elif result_byte == 0x77:
        raise InvalidParameterError
    elif result_byte == 0x88:
        raise DeviceDisabledError("Device is disabled")
    elif result_byte == 0xDD:
        raise HealthTestError()
    elif result_byte == 0xAA:
        pass # do nothing, command was successful
    else:
        raise UnknownResultByteError('Unknown result byte 0x{:02X}'.format(result_byte))

class DS28C50:
    def __init__(self, i2c):
        self.adapter = i2c

    _address = 0x38 #write address

    class _CmdDelay:
        tRM = 50
        tWM = 100
        tWM_BL = 60
        tWS = 60
        tCMP = 5
        tRNG = 25
        tODC = 50
        
    class PgProt:
        RP   = 0b00000001
        WP   = 0b00000010
        EM   = 0b00000100
        APH  = 0b00001000
        EPH  = 0b00010000
        NONE = 0b00100000
        EF   = 0b01000000
    
    class DisableOp:
        SetRelease = 0b00001111
        LockoutCommand = 0b00000101
        DisableDevice = 0b00000000
    
    class SecretParam:
        pg6 =  0b00
        secret_s = 0b10
    
    def WriteMemory(self, page, write_data):
        '''
        DS28C50 Write Memory Command (0x96) - Writes 32 bytes of data to device memory page.

        Keyword Arguments:
            page (int) - page to write to. Valid values 0 to 8.
            write_data (list[int]) - 32 bytes of data to write to page.

        Returns:
            Device Result Byte (int):
            0xAA = success
            0x55 = command failed due to page protection
            0x77 = invalid input or parameter
            0x88 = device is disabled

        Command Restrictions:
            Target page must not have: WP, APH, or EPH protection

        See DS28C50 User's Guide pg 14 for more info.
        '''
        
        data = []
        data.append(page)
        data.extend(write_data)
        result = self.adapter.StdCmdFlow(DS28C50._address, 0x96, data, self._CmdDelay.tWM) #i2c address, cmd, writebuf, delay
        #return result[0]
        processResultByte(result[0])
    
    def WriteMemoryEx(self, page, write_data):
        '''
        DS28C50 Write Memory EX Command (0x97) - Writes 32 bytes of data to device memory
            page, but will only write to blocks that need to be changed in page.

        Keyword Arguments:
            page (int) - page to write to. Valid values 0 to 8.
            write_data (list[int]) - 32 bytes for data to write to page.

        Returns:
        Device Result Byte (int):
            0xAA = success
            0x55 = command failed due to page protection
            0x77 = invalid input or parameter
            0x88 = device is disabled
        Bitmap of written blocks (int):
            bit0 = 1 (block0 written)
            bit1 = 1 (block1 written)
            bit2 = 1 (block2 written)
            bit3 = 1 (block3 written)
        
        Command Restricitons:
            Target page must not have: WP, APH, or EPH protection

        See DS28C50 User's Guide pg 15-16 for more info.
        '''
        
        data = []
        data.append(page)
        data.extend(write_data)
        result = self.adapter.StdCmdFlow(DS28C50._address, 0x97, data, self._CmdDelay.tWM)
        
        processResultByte(result[0])
        return result[1]
        #if(len(result) > 1):
        #    return result[0], result[1]
        #else:
        #    return result[0], None

    def ReadMemory(self, page):
        '''
        DS28C50 Read Memory Command (0x44) - Reads page data (32 bytes) from device memory.

        Keyword Arguments:
        page (int) - page to read. Valid values 0 to 8

        Returns:
        Device Result Byte (int)
            0xAA = Success
            0x55 = Page is read protected
            0x77 = Invalid input or parameter
            0x88 = Device disabled
        Page Read Data (list[int]) - 32 bytes of page data

        Command Restricitons:
        Target page must not have RP, EPH

        See DS28C50 User's guide pg 17 for more info.
        '''
        
        result = self.adapter.StdCmdFlow(DS28C50._address, 0x44, [page], self._CmdDelay.tRM)
        processResultByte(result[0])
        return result[1:]

    def ReadStatus(self):
        '''
        DS28C50 Read Status Command (0xAA) - Reads Protection state of all pages

        Keyword Arguments:
        None

        Returns:
        Device Result Byte (int)
            0xAA = Success
            0x88 = Device disabled
        Page protections (list[int])
            One byte per page.
            bit0 = 1 (RP enabled)
            bit1 = 1 (WP enabled)
            bit2 = 1 (EM enabled)
            bit3 = 1 (APH enabled)
            bit4 = 1 (EPH enabled)
            bit5 = 1 (NONE enabled)
            bit6 = 1 (EF enabled)

        Command Restritions: None

        See DS28C50 User's guide page 18
        '''
        
        result = self.adapter.StdCmdFlow(DS28C50._address, 0xAA, [], self._CmdDelay.tRM)
        processResultByte(result[0])
        return result[1:]

    def SetPageProtection(self, page, protection):
        '''
        DS28C50 Set Page Protection Command (0xC3) - Sets protection state to selected page

        Keyword Arguments:
        page (int) - page to set protecion. Valid values 0 to 8.
        protection (int) - protections to set on page.
            bit0 = 1 (RP enabled)
            bit1 = 1 (WP enabled)
            bit2 = 1 (EM enabled)
            bit3 = 1 (APH enabled)
            bit4 = 1 (EPH enabled)
            bit5 = 1 (NONE enabled)
            bit6 = 1 (EF enabled)

        Returns:
        Device Result Byte (int)
            0xAA = Success
            0x77 = Invalid parameter combination
            0x55 = The command failed because protection for the page was already done
            0x88 = Device disabled (result length 1)
            
        Command Restrictions:
        Page must not already have protections set.
        protection must be valid for page. (see table 2 on page 8 in DS28C50 User's guide)

        See DS28C50 User's guide pages 7-8, 19-20 for more details
        '''

        delay = self._CmdDelay.tWS

        if(page == 5):
            delay += self._CmdDelay.tWM
            
        result = self.adapter.StdCmdFlow(DS28C50._address, 0xC3, [page, protection], delay)
        processResultByte(result[0])
        #return result[0]

    def ReadRng(self, num_bytes, eht=False):
        '''
        DS28C50 Read RNG Command (0xD2) - Compute and read random data from the true random number generator
        
        Keyword Arguments:
        num_bytes (int) - number of random bytes to read - 1. Valid values 0 to 63.
        eht (bool) - run entropy health test and ignore num_bytes.
        
        Returns:
        Device Result Byte (int)
            AAh = Success
            77h = Invalid parameter
            DDh = Entropy health test fail (only possible if EHT = 1)
            22h = Failure to generate random number
            88h = Device disabled
        Random Data (list[int]) - list of random bytes of length num_bytes + 1 or None if eht=True
        
        Command Restricitons:
        Entropy health test option takes precedence over RNG read.
        
        See DS28C50 User's guide page 21 for more details
        '''
        
        param = (0xFF & num_bytes)
        delay = _CmdDelay.tRNG
        if(eht):
            param = param | 0x80 #set bit 7 to enable eht
            delay = _CmdDelay.tODC
            
        result = self.adapter.StdCmdFlow(DS28C50._address, 0xD2, [param], delay)
        
        processResultByte(result[0])
        if(eht):
            return None
        else:
            return result[1:]
        
    def EncryptedReadMemory(self, page, secret_param):
        '''
        DS28C50 Encrypted Read Memory Command (0x4B) - Performs an encrypted read of a protected page or special function.
        
        Keyword Arguments:
        page (int) - page to read. Valid values 0 to 5, 7, 8.
        secret_param (int) - secret to use
            0b00 = use page 6 secret
            0b10 = use secret_s
            
        Returns:
        Device Result Byte (int)
            AAh = Success
            55h = The command failed because the destination page has the wrong protection; requires EPH 
            77h = Invalid input or parameter 
            88h = Device disabled
        Challenge (list[int]) - list of 8 bytes used to calculate decryption hmac
        Encrypted Data (list[int]) - list of 32 bytes of encrypted page data
        
        Command Restricitons:
        This command is applicable only to memory pages that do not have read protection and have encryption protection. 
        Reading a read-protected page including pages that have intrinsic read protection, like Secret, causes FFh to be read in all bytes, including the challenge.
        
        See DS28C50 User's guide pages 22-23 for more details
        '''
        
        cmd_parameters = []
        cmd_parameters.append(page & 0x00FF)
        cmd_parameters.append(secret_param & 0x03)
        
        result = self.adapter.StdCmdFlow(DS28C50._address, 0x4B, cmd_parameters, self._CmdDelay.tCMP+self._CmdDelay.tRM)
        
        processResultByte(result[0])
        return result[1:9], result[9:]

    def ComputeAndReadPageAuthentication(self, page, auth_type, challenge):
        '''
        Compute and Read Page Authentication (0xA5) - Computes and returns a 32bit HMAC using SHA-3-256 for selected page.

        Keyword Arguments:
        page (int) - page to use to compute hmac. Valid values 0 to 8.
        auth_type - selects wich secret to use
            0 = use page 6 secret
            2 = use computed secret S

        Returns:
        Device Result Byte (int)
            0xAA = Success
            0x22 = Failure to create signature
            0x33 = Invalid sequence; required step not done (RMR not set and required)
            0x77 = Invalid input or parameter
            0x88 = Device disabled

        HMAC (list[int]) - sha3 hmac calculated by compute and read page authentication (32 bytes) 
            
        Command Restrictions:
        RMR must be set prior to using this command on (ROM Options) or (GPIO Control).
        RMR is set using either the Read Memory command or the Encrypted Read Memory command.

        See DS28C50 User's guide pages 24-27 for more details.
        '''

        cmd_parameters = []
        cmd_parameters.append(page)
        cmd_parameters.append(auth_type)
        cmd_parameters.extend(challenge)
        
        result = self.adapter.StdCmdFlow(DS28C50._address, 0xA5, cmd_parameters, self._CmdDelay.tCMP+self._CmdDelay.tRM)
        processResultByte(result[0])
        return result[1:]

    def AuthenticatedSha3WriteMemory(self, page, secret_param, write_data, hmac, challenge=[]):
        '''
        DS28C50 Authenticated SHA-3 Write Memory Command (0x99) - The Authenticated SHA-3 Write Memory command writes with authentication using HMAC SHA-3-256 with optional encryption.
        
        Keyword Arguments:
        page (int) - page to write. valid values 0 - 5, 7, 8.
        secret_param (int) - secret to use
            0b00 = use page 6 secret
            0b10 = use secret_s
        write_data (list[int]) - 32 bytes of page data (or encrypted data) to write.
        hmac - SHA-3 hmac calculated using message specified by table 30 (page 31) in user's guide
        challenge (list[int]) - (required only if page has eph protecion) 8 byte challenge used to calculate decryption hmac 
        
        Returns:
        Device Result Byte (int)
            AAh = Success
            33h = Invalid sequence, required step not done (RMR not set and required)
            55h = Invalid Protection, not APH or EPH
            77h = Invalid input or parameter
            00h = Invalid HMAC write authorization
            88h = Device disabled
            
        Command Restricitons:
        This command is applicable only to memory pages that already have master authentication set (APH or EPH). 
        The HMAC is computed with Old Page Data that is read from the page. 
        If the page is ROM Options, this could include values of FFh for the ROM ID when in anonymous mode. 
        RMR must be set prior to using this command on page (ROM Options) or page (GPIO Control). 
        RMR is set using either Read Memory or Encrypted Read Memory. 
        This requirement loads the RANDOM_PAGE_SEED so that each call to Authenticated SHA-3 Write Memory must require a different HMAC. 
        At the end of this function RMR is cleared for all pages.
        
        See DS28C50 User's guide pages 28 - 31 for more details
        '''
        
        delay = 0
        if(len(challenge) > 0):
            delay = (self._CmdDelay.tCMP*2) + self._CmdDelay.tWM
        else:
            delay = self._CmdDelay.tCMP + self._CmdDelay.tWM
    
        cmd_parameters = []
        cmd_parameters.append(page & 0x00FF)
        cmd_parameters.append(secret_param & 0x03)
        cmd_parameters.extend(write_data)
        cmd_parameters.extend(hmac)
        cmd_parameters.extend(challenge)
        
        result = self.adapter.StdCmdFlow(DS28C50._address, 0x99, cmd_parameters, delay)
        #return result[0]
        processResultByte(result[0])
        
    def AuthenticatedSha3WriteMemoryEx(self, page, secret_param, write_data, hmac, challenge=[]):
        '''
        DS28C50 Command Authenticated SHA-3 Write Memory EX (0x) - Similar to the Authenticated SHA-3 Write Memory command, the Authenticated SHA-3 Write Memory EX command writes with authentication using HMAC SHA-3-256 with optional encryption. However, upon completion, the Write Memory EX command returns a bitmap indicating which of the four 8-byte blocks have been written in a 32-byte page write.
        
        Keyword Arguments:
        page (int) - page to write. valid values 0 - 5, 7, 8.
        secret_param (int) - secret to use
            0b00 = use page 6 secret
            0b10 = use secret_s
        write_data (list[int]) - 32 bytes of page data (or encrypted data) to write.
        hmac - SHA-3 hmac calculated using message specified by table 30 (page 31) in user's guide
        challenge (list[int]) - (required only if page has eph protecion) 8 byte challenge used to calculate decryption hmac 
        
        Returns:
        Device Result Byte (int)
            AAh = Success
            33h = Invalid sequence, required step not done (RMR not set and required)
            55h = Invalid Protection, not APH or EPH
            77h = Invalid input or parameter
            00h = Invalid HMAC write authorization
            88h = Device disabled
        Bitmap (int)
            Bitmap of written blocks (int):
            bit0 = 1 (block0 written)
            bit1 = 1 (block1 written)
            bit2 = 1 (block2 written)
            bit3 = 1 (block3 written)
        
        Command Restricitons:
        This command is applicable only to memory pages that already have master authentication set (APH or EPH). 
        The HMAC is computed with Old Page Data that is read from the page. 
        If the page is ROM Options, this could include values of FFh for the ROM ID when in anonymous mode. 
        RMR must be set prior to using this command on page (ROM Options) or page (GPIO Control). 
        RMR is set using either Read Memory or Encrypted Read Memory. 
        This requirement loads the RANDOM_PAGE_SEED so that each call to Authenticated SHA-3 Write Memory must require a different HMAC. 
        At the end of this function RMR is cleared for all pages.
        
        See DS28C50 User's guide pages 32 - 33 for more details
        '''
        
        delay = 0
        if(len(challenge) > 0):
            delay = (self._CmdDelay.tCMP*2) + self._CmdDelay.tWM
        else:
            delay = self._CmdDelay.tCMP + self._CmdDelay.tWM
    
        cmd_parameters = []
        cmd_parameters.append(page & 0x00FF)
        cmd_parameters.append(secret_param & 0x03)
        cmd_parameters.extend(write_data)
        cmd_parameters.extend(hmac)
        cmd_parameters.extend(challenge)
        
        result = self.adapter.StdCmdFlow(DS28C50._address, 0x9A, cmd_parameters, delay)
        
        processResultByte(result[0])
        return result[1]


    def ComputeAndLockSha3Secret(self, binding_data_pg, wpe, dsecret, msecret, partial_secret):
        '''
        DS28C50 Compute and Lock SHA-3 Secret Command (0x3C) - Computes a secret based on a partial secret, page data, and the current secret.
        
        Keyword Arguments:
        binding_data_pg (int) - page to use binding data. Valid values 0 to 5.
        wpe (bool) - set WP protection on secret
        dsecret (int) - destination secret location
            0b00 = use page 6
            0b10 = use secret_s
        msecret (int) - source secret used sha3 hmac calculation
            0b00 = use page 6 secret
            0b10 = use secret_s
            
        partial_secret (list[32]) - 32 bytes of data used in sha3 hmac calculation
        
        Returns:
        Device Result Byte (int)
            AAh = Success 
            55h = The command failed because the destination secret is protected 
            77h = Invalid input or parameter 
            88h = Device disabled
        
        Command Restricitons:
        Destination secret must not have been written already.
        
        See DS28C50 User's guide pages 34 - 36 for more details
        '''
        
        prot_sec_param = 0x00
        if(wpe):
            prot_sec_param |= (0x40)

        prot_sec_param |= ((dsecret & 0x03) << 2)        
        prot_sec_param |= (msecret & 0x03)
        
        cmd_parameters = []
        cmd_parameters.append(binding_data_pg & 0x00FF)
        cmd_parameters.append(prot_sec_param)
        cmd_parameters.extend(partial_secret)
        result = self.adapter.StdCmdFlow(DS28C50._address, 0x3C, cmd_parameters, self._CmdDelay.tCMP+self._CmdDelay.tWM)
        #return result[0]
        processResultByte(result[0])
        
    def DecrementCounter(self):
        '''
        DS28C50 Command Decrement Counter (0x96) - The Decrement Counter command is used to decrement the decrement-only counter.
        
        Keyword Arguments:
        None
        
        Returns:
        Device Result Byte (int)
            AAh = Success
            55h = The command failed because the counter is already 0
            33h = The command failed because the counter was never set
            22h = Decrement Counter value not valid
            88h = Device disabled
            
        Command Restricitons:
        Requires Page 5 to have EF (enhanced protection).
        
        See DS28C50 User's guide page 37 for more details
        '''
        
        cmd_parameters = []
        result = self.adapter.StdCmdFlow(DS28C50._address, 0xC9, cmd_parameters, self._CmdDelay.tWM)
        #return result[0]
        processResultByte(result[0])
        
    def DeviceDisable(self, operation, release_sequence):
        '''
        DS28C50 Device Disable Command (0x33) - Command to permanently disable the device. 
        This command is only available if Enhanced Function (EF) is enabled on Page 5. 
        The disable sequence or password is set once EF protection is enabled.
        
        Keyword Arguments:
            operation (int) - determains what function of device disable is run
                0b00001111 = Set Release value
                0b00000101 = Lock out device disable command
                0b00000000 = Disable device
                
        Returns:
        Device Result Byte (int)
        AAh = Success 
        77h = Invalid parameter 
        33h = Bad operation (attempting to set password if password already set etc.) 
        00h = Release sequence is incorrect 
        88h = Device disabled
        
        Command Restricitons:
        EF protection must be set on Page 5.
        
        See DS28C50 User's guide page 38 for more details
        '''
        
        cmd_parameters = []
        cmd_parameters.append(operation & 0xFF)
        cmd_parameters.extend(release_sequence)
        result = self.adapter.StdCmdFlow(DS28C50._address, 0x33, cmd_parameters, self._CmdDelay.tWS)
        #return result[0]
        processResultByte(result[0])