import serial
import serial.tools.list_ports
import time

class DS9481P:
    def __init__(self):
        self.Connected = False
        self.ComPort = None

    def Open(self, com, timeoutVal = 2.0):
        if (self.ComPort == None):
            self.ComPort = serial.Serial(port=com, timeout=2.0)
        elif(not self.ComPort.is_open):
            self.ComPort.port = com
            self.ComPort.baudrate = 9600
            self.ComPort.open()
        self.ComPort.break_condition = True
        time.sleep(0.002) #wait for 2 ms
        self.ComPort.break_condition = False
        self.Connected = self.ComPort.is_open

    def Close(self):
        self.ComPort.close()
        self.Connected = self.ComPort.is_open

    def WriteCom(self, data):
        for i in range(len(data)):
            try:
                data[i] = ord(data[i])
        
            except TypeError:
                if(isinstance(data[i], int)):
                    pass
                else:
                    raise
        try:
            self.ComPort.write(bytearray(data))
            
        except serial.serialutil.SerialException as e:
            if (not self.ComPort.is_open):
                print(e)
                raise
        except AttributeError:
            print("No serial object: Call Open before using WriteCom")
            
    def ReadCom(self, length=1):
        return self.ComPort.read(length)
        

class I2C:
    def __init__(self, ds9481p = None):
        if(ds9481p is None):
            ds9481p = DS9481P()
        self.Adapter = ds9481p
        
    def Connect(self, com):
        if(not self.Adapter.Connected):
            self.Adapter.Open(com=com)
        self.Adapter.WriteCom([0xC1]) #send DS2480 sync byte
        self.Adapter.WriteCom([0xE5]) #exit to I2C mode
        awake = self.Adapter.ReadCom()#read awake byte from I2C mode
        
    def Disconnect(self):
        self.Adapter.Close()

    def Start(self):
        self.Adapter.WriteCom(['S'])

    def Stop(self):
        self.Adapter.WriteCom(['P'])

    def RepeatedStart(self):
        self.Adapter.WriteCom(['T'])

    def WriteByte(self, byte, expectAck):
        if(expectAck):
            self.Adapter.WriteCom(['Q'])
            self.Adapter.WriteCom([byte])
            val = self.Adapter.ReadCom()
            return (bytearray(val)[0] & 0x01) == 0x00
        else:
            self.Adapter.WriteCom(['W'])
            self.Adapter.WriteCom([byte])

    def ReadByte(self, sendAck=True):
        if(sendAck):
            self.Adapter.WriteCom(['R'])
        else:
            self.Adapter.WriteCom(['N'])
        return bytearray(self.Adapter.ReadCom())[0]

    def Packet(self, flavor, data, readLen=0):
        self.Adapter.WriteCom(['Z'])
        self.Adapter.WriteCom([flavor])
        self.Adapter.WriteCom([len(data)])
        
        if(flavor == 0x01): #write only
            self.Adapter.WriteCom(data)
            self.Adapter.ReadCom()#read crc and discard
            return
        
        elif(flavor == 0x02): #write and read
            self.Adapter.WriteCom(data)
            self.Adapter.WriteCom(readLen)
            read_data = []
            for i in range(readLen):
                read_data.append(self.Adapter.ReadCom())
            self.Adapter.ReadCom() #read crc and discard
            return read_data

        #TODO: add support for rest of packet command
        

    def StdCmdFlow(self, address, cmd, writebuf, delay):
        read_data = []
        ms_delay = float(delay)/1000.0
        tx_data = [len(writebuf)]
        tx_data.extend(writebuf)
        
        self.Start()
        #send write address
        if(not self.WriteByte(address, True)):
            self.Start()
            if(not self.WriteByte(0x00, True)):
                self.Stop()
                return [0x01, 0xF0]
            
        #send command byte
        if(not self.WriteByte(cmd, True)):
            self.Stop()
            return [0x01, 0xF1]
            
        #send parameters and data if any
        if(len(writebuf) > 0):
            for byte in tx_data:
                if(not self.WriteByte(byte, True)):
                    self.Stop()
                    return [0x01, 0xF2]
                
        self.Stop()
        
        time.sleep(ms_delay)

        self.Start()
        if(not self.WriteByte((address | 0x01), True)):
                self.Stop()
                return [0x01, 0xF3]

        read_len = self.ReadByte()

        ack = True
        for i in range(read_len):
            if(i == (read_len - 1)):
                ack = False
            else:
                ack = True
            read_data.append(self.ReadByte(ack))

        self.Stop()
        
        return read_data
